﻿#include <windows.h>
#include <iterator>
#include <utility>
#include <memory>
#include <string>
#include <vector>

#include "MQPlugin.h"

#ifndef _CMQObjectDefine
#define _CMQObjectDefine

namespace cmq {

typedef int oindex_t;
typedef int vindex_t;
typedef int findex_t;
typedef int64_t lindex_t;
typedef int cindex_t;
typedef int mindex_t;

lindex_t makeLineId(vindex_t v1, vindex_t v2);
lindex_t makeLineId(std::pair<vindex_t, vindex_t> line);
std::pair<vindex_t, vindex_t> getLineVertex(lindex_t lindex);

class CMQPoint {
public:
	static float innercross(MQPoint p1, MQPoint p2) {
		return p1.x * p2.x + p1.y * p2.y + p1.z * p2.z;
	}

	static MQPoint outercross(MQPoint p1, MQPoint p2) {
		return MQPoint(p1.y * p2.z - p1.z * p2.y, p1.z * p2.x - p1.x * p2.z, p1.x * p2.y - p1.y * p2.x);
	}
};

class CMQObject;

class CMQVertex {
public:
	CMQVertex(const MQPoint &p, CMQObject *obj, vindex_t index) : m_p(p), m_obj(obj), m_index(index) {}
	~CMQVertex() {
		m_index = -1;
		m_obj = nullptr;
		m_p.zero();
	}
	const MQPoint &p() const {return m_p;}
	bool isSelect() const;
	void addSelect() const;
	void deleteSelect() const;
	std::vector<findex_t> getVertexRelatedFaces() const;

	vindex_t m_index;
	CMQObject *m_obj;
	MQPoint m_p;
};

class CMQFace {
public:
	CMQFace(CMQObject *obj, findex_t index) : m_obj(obj), m_index(index) {}
	~CMQFace() {
		m_obj = nullptr;
		m_index = -1;
	}
	bool isSelect() const;
	void addSelect() const;
	void deleteSelect() const;
	int getFacePointCount() const;
	std::vector<vindex_t> getFacePointArray() const;
private:
	findex_t m_index;
	CMQObject *m_obj;
};

class CMQCoordinate
{
public:
	CMQCoordinate() : index_(-1) {cood_.u =cood_.v = 0.0f;}
	CMQCoordinate(CMQCoordinate &&cood) : index_(cood.index_), cood_(cood.cood_) {}
	CMQCoordinate &operator=(CMQCoordinate &&cood) {index_ = cood.index_; cood_ = cood.cood_; return *this;}

	CMQCoordinate(cindex_t index, const MQCoordinate &cood) : index_(index), cood_(cood) {}
	bool operator==(const CMQCoordinate &other) const {return index_ == other.index_;}
	static bool uvcomp(const CMQCoordinate &left, const CMQCoordinate &right) {
		return pow2(left.cood_.u - right.cood_.u) + pow2(left.cood_.v - right.cood_.v) < pow2(uvdelta);
	}
	cindex_t getIndex() const {return index_;}
	MQCoordinate getCoordinate() const {return cood_;}
private:
	cindex_t index_;
	MQCoordinate cood_;

	static const float uvdelta; // UV値同一判定用
	static float pow2(float v) {return v * v;}
};

template <class T>
class CMQIterator : public std::iterator<std::bidirectional_iterator_tag, int> {
public:
	friend class CMQObject;

	CMQIterator(const CMQIterator &iter) :
		m_container(iter.m_container), m_index(iter.m_index) {}

	CMQIterator &operator++() {
		m_index = next(m_index, true);
		return *this;
	}
	const CMQIterator &operator++() const {return const_cast<CMQIterator<T>*>(this)->operator++();}
	CMQIterator operator++(int) {
		CMQIterator origiter(*this);
		this->operator++();
		return origiter;
	}
	const CMQIterator operator++(int) const {reuturn const_cast<CMQIterator<T>*>(this)->operator++(0);}

	CMQIterator &operator--() {
		m_index = next(m_index, false);
		return *this;
	}
	const CMQIterator &operator--() const {return const_cast<CMQIterator<T>*>(this)->operator--();}
	CMQIterator operator--(int) {
		CMQIterator<T> origiter(*this);
		this->operator--();
		return origiter;
	}
	const CMQIterator operator--(int) const {reuturn const_cast<CMQIterator<T>*>(this)->operator--(0);}

	bool operator!=(const CMQIterator &iter) const {return m_index != iter.m_index;}
	bool operator==(const CMQIterator &iter) const {return !(this->operator!=(iter));}

//	const T &operator*() const {return get();}
//	T &operator*() {return const_cast<T&>(get());}
	std::shared_ptr<T> operator->() const {
		std::shared_ptr<T> ptr(getp());
		return ptr;
	}
//	T *operator->() {return &get(m_index);}

	int index() const {return m_index;}
	const T &get() const;
	std::shared_ptr<T> getp() const;

private:
	int m_index;
	CMQObject *m_container;

	CMQIterator() : m_container(NULL), m_index(-1) {}
	CMQIterator(CMQObject * const container, int index) :
		m_container(container), m_index(index) {}

	int getCount() const {return m_container->getCount<T>();}
	bool isValid(int) const {return m_container->isValid<T>(index);}
	int next(int index, bool forward) const {return m_container->next<T>(index, forward);}
};

class CMQObject
{
friend class CMQDocument;
friend class Vertex;
friend class Face;
public:
	CMQObject(MQObject obj, CMQDocument *doc) : m_obj(obj), m_doc(doc) {}
	CMQObject(CMQObject &&obj) : m_obj(obj.m_obj), m_doc(obj.m_doc) {}
	CMQObject() : m_obj(nullptr), m_doc(nullptr) {}
	CMQObject &operator=(CMQObject &&obj) {m_obj = obj.m_obj; m_doc = obj.m_doc; return *this;}

	void setDocument(CMQDocument *doc) {m_doc = doc;}
	operator bool() const {return m_obj != nullptr;}
	CMQObject clone();
	std::string getName();
	void setName(const std::string &name);
	UINT getUniqueID() {return m_obj->GetUniqueID();}
	bool getSelected() {return m_obj->GetSelected() == TRUE;}
	bool getLocking() {return m_obj->GetLocking() == TRUE;}
	bool getVisible() {return m_obj->GetVisible() == TRUE;}
	bool getFolding() {return m_obj->GetFolding() == TRUE;}
	void setSelected(bool flag) {m_obj->SetSelected(flag);}
	bool setLocking(bool flag) {m_obj->SetLocking(flag);}
	bool setVisible(bool flag) {m_obj->SetVisible(flag);}
	bool setFolding(bool flag) {m_obj->SetFolding(flag);}
	int getDepth() {return m_obj->GetDepth();}
	void setDepth(int depth) {m_obj->SetDepth(depth);}
	void deleteThis() {m_obj->DeleteThis();}

	int getVertexCount() const {return m_obj->GetVertexCount();}
	int getVertexRefCount(vindex_t vindex) const {return m_obj->GetVertexRefCount(vindex);}
	MQPoint getVertex(vindex_t vindex) const {return m_obj->GetVertex(vindex);}
	UINT getVertexUniqueID(vindex_t vindex) const {return m_obj->GetVertexUniqueID(vindex);}
	vindex_t getVertexIndexFromUniqueID(UINT uid) const {return m_obj->GetVertexIndexFromUniqueID(uid);}
	void setVertex(vindex_t vindex, MQPoint p) {m_obj->SetVertex(vindex, p);}
	vindex_t addVertex(const MQPoint &p) {return m_obj->AddVertex(p);}
	vindex_t addVertex(const CMQVertex &v) {return addVertex(v.p());}
	BOOL deleteVertex(vindex_t index) {return m_obj->DeleteVertex(index);}
	bool isSelectVertex(vindex_t index);
	void addSelectVertex(vindex_t index);
	void deleteSelectVertex(vindex_t index);
	void copyVertexAttribute(vindex_t vert1, CMQObject obj2, vindex_t vert2) {m_obj->CopyVertexAttribute(vert1, obj2.m_obj, vert2);}

	int getFaceCount() const {return m_obj->GetFaceCount();}
	int getFacePointCount(findex_t findex) const {return m_obj->GetFacePointCount(findex);}
	UINT getFaceUniqueID(findex_t findex) const {return m_obj->GetFaceUniqueID(findex);}
	bool getFaceVisible(findex_t findex) const {return m_obj->GetFaceVisible(findex) == TRUE;}
	void setFaceVisible(findex_t findex, bool visible) {m_obj->SetFaceVisible(findex, visible ? TRUE: FALSE);}
	findex_t addFace(std::vector<vindex_t> varray);
	BOOL deleteFace(findex_t index) {return m_obj->DeleteFace(index);}
	int modifyFace(findex_t index, const std::vector<vindex_t> &varray);
	std::vector<findex_t> getVertexRelatedFaces(vindex_t vindex);
	std::vector<vindex_t> getFacePointArray(findex_t findex);
	mindex_t getFaceMaterial(findex_t findex) const {return m_obj->GetFaceMaterial(findex);}
	void setFaceMaterial(findex_t findex, mindex_t mindex) {m_obj->SetFaceMaterial(findex, mindex);}
	std::vector<CMQCoordinate> getFaceCoordinateArray(findex_t findex);
	void setFaceCoordinate(findex_t findex, vindex_t vertex, const CMQCoordinate &co);
	void setFaceCoordinateArray(findex_t findex, const std::vector<CMQCoordinate> &carray);
	DWORD getFaceVertexColor(findex_t findex, findex_t vindex) const {return m_obj->GetFaceVertexColor(findex, vindex);}
	void setFaceVertexColor(findex_t findex, findex_t vindex, DWORD color) {m_obj->SetFaceVertexColor(findex, vindex, color);}
	float getFaceEdgeCrease(findex_t findex, int lindex) const {return m_obj->GetFaceEdgeCrease(findex, lindex);}
	void setFaceEdgeCrease(findex_t findex, int lindex, float crease) {m_obj->SetFaceEdgeCrease(findex, lindex, crease);}
	bool isSelectFace(findex_t index);
	void addSelectFace(findex_t index);
	void deleteSelectFace(findex_t index);

	std::pair<vindex_t, vindex_t> getLineVertex(findex_t findex, int lindex);
	MQPoint getLineVector(findex_t findex, int lindex);
	MQPoint getNormalVector(findex_t findex, int lindex = 0);
	MQPoint getNeiborPoint(findex_t findex, int lindex);
	findex_t createLine(MQPoint &p1, MQPoint &p2);
	std::vector<std::pair<findex_t, int>> getOtherLine(vindex_t v1, vindex_t v2, findex_t face = -1);
	std::vector<std::pair<findex_t, int>> getOtherLine(std::pair<vindex_t, vindex_t> line, findex_t face = -1);
	std::vector<lindex_t> getLineArray(findex_t findex, vindex_t vindex = -1);

	template <class T> int getCount() const;
	template <class T> bool isValid(int index) const;
//	template <class T> T get(int index) const;

	template <class T> int next(int index, bool forward) {
		int count = getCount<T>();
		do {
			if(forward) index++; else index--;
		} while(index >= 0 && index < count && !isValid<T>(index));
		return index;
	}

	template <class T> CMQIterator<T> begin(){
		int index = 0;
		while(index < getCount<T>() && !isValid<T>(index)) {
			index = next<T>(index, true);
		}
		return CMQIterator<T>(this, index);
	}
	template <class T> const CMQIterator<T> cbegin() const {
		return const_cast<CMQObject*>(this)->begin();
	}
	template <class T> CMQIterator<T> end() {
		CMQIterator<T> iter(this, getCount<T>());
		return iter;
	}
	template <class T> const CMQIterator<T> cend() const {
		return const_cast<CMQObject*>(this)->end();
	}

	//private:
	MQObject m_obj;
	CMQDocument *m_doc;
};

class CMQDocument
{
friend class CMQObject;
friend class Vertex;
friend class Face;
public:
	CMQDocument(MQDocument doc) : m_doc(doc) {}
	CMQDocument(CMQDocument &&doc) : m_doc(doc.m_doc) {}
	CMQDocument &operator=(CMQDocument &&doc) {m_doc = doc.m_doc; return *this;}

	CMQObject getCurrentObject();
	CMQObject &insertObject(CMQObject &newobj, const CMQObject &before);
	void deleteObject(oindex_t i) {m_doc->DeleteObject(i);}
	void deleteObject(CMQObject obj) {oindex_t i = getObjectIndex(obj); if(i == -1) obj.m_obj->DeleteThis(); else m_doc->DeleteObject(i);}
	std::string getUnusedName(const std::string &name);
	int getObjectCount() {return m_doc->GetObjectCount();}
	CMQObject getObject(oindex_t i) {return CMQObject(m_doc->GetObject(i), this);}
	oindex_t getObjectIndex(CMQObject obj) {return m_doc->GetObjectIndex(obj.m_obj);}
	CMQObject getObjectbyName(const std::string &name);

	mindex_t getCurrentMaterialIndex() {return m_doc->GetCurrentMaterialIndex();}
	int getMaterialCount() {return m_doc->GetMaterialCount();}
	MQMaterial getMaterial(mindex_t i) {return MQMaterial(m_doc->GetMaterial(i));}
	void addMaterial(const MQMaterial &m) {m_doc->AddMaterial(m);}

	bool isSelectVertex(CMQObject obj, vindex_t vindex) {return m_doc->IsSelectVertex(m_doc->GetObjectIndex(obj.m_obj), vindex) == TRUE;}
	bool addSelectVertex(CMQObject obj, vindex_t vindex) {return m_doc->AddSelectVertex(m_doc->GetObjectIndex(obj.m_obj), vindex) == TRUE;}
	bool deleteSelectVertex(CMQObject obj, vindex_t vindex) {return m_doc->DeleteSelectVertex(m_doc->GetObjectIndex(obj.m_obj), vindex) == TRUE;}

	bool isSelectFace(CMQObject obj, findex_t findex) {return m_doc->IsSelectFace(m_doc->GetObjectIndex(obj.m_obj), findex) == TRUE;}
	bool addSelectFace(CMQObject obj, findex_t findex) {return m_doc->AddSelectFace(m_doc->GetObjectIndex(obj.m_obj), findex) == TRUE;}
	bool deleteSelectFace(CMQObject obj, findex_t findex) {return m_doc->DeleteSelectFace(m_doc->GetObjectIndex(obj.m_obj), findex) == TRUE;}

	bool isSelectLine(CMQObject obj, findex_t findex, int lindex) {return m_doc->IsSelectLine(m_doc->GetObjectIndex(obj.m_obj), findex, lindex) == TRUE;}
	bool addSelectLine(CMQObject obj, findex_t findex, int lindex) {return m_doc->AddSelectLine(m_doc->GetObjectIndex(obj.m_obj), findex, lindex) == TRUE;}
	bool deleteSelectLine(CMQObject obj, findex_t findex, int lindex) {return m_doc->DeleteSelectLine(m_doc->GetObjectIndex(obj.m_obj), findex, lindex) == TRUE;}

	void clearSelect(DWORD flag) {m_doc->ClearSelect(flag);}

private:
	MQDocument m_doc;
};

template <class T, class P>
class CMQContainer
{
public:
	CMQContainer(P *pobj) : m_pobj(pobj){}

	CMQIterator<T> begin() {return m_pobj->begin<T>();}
	const CMQIterator<T> begin() const {return m_pobj->cbegin<T>();}
	CMQIterator<T> end() {return m_pobj->end<T>();}
	const CMQIterator<T> end() const {return m_pobj->cend<T>();}

private:
	P *m_pobj;
};

std::wstring towstring(const std::string &s);

std::string tostring(const std::wstring &s);

}
#endif
